/** <module> db high level SQL database access module.
	Ported from XSB see.

	@license GPL
*/

:-module(db, [
		     db_open/1, db_open/2, db_open/3, db_open/4,
		     db_close/0, db_close/1,

		     db_create_table/2, db_create_table/3,
		     db_delete_table/1, db_delete_table/2,

		     %db_create_index/4, db_create_index/5,
		     %db_delete_index/2, db_delete_index/3,
		     %db_delete_view/1, db_delete_view/2,
		     db_sql/3, db_sql/4,
		     findall_db_sql/3, findall_db_sql/4,

		     db_import/2, db_import/3, db_import/4,

		     db_insert/2, db_insert/3,
		     db_delete/2, db_delete/3,


		     db_query/2, db_query/3, db_query/4,
		     db_deallocate_query/1, db_deallocate_query/2,

		     db_update/2, db_update/3, db_update/4,

		     db_flag/2, set_db_flag/2
		    ]).


%%%%%%%%	some swi specific stuff	%%%%%%%%
%%%%
%%%%
%%%%

%%	term_to_elements(?Term, ?M, ?Functor, ?Args)
%
%	Mode advanced version of =..
%
term_to_elements(Term, M, Term, []):-
	atomic(Term),!,
	strip_module(Term,M,_Plain).
term_to_elements(Term, M, Functor, Args):-
	compound(Term),
	strip_module(Term, M, Plain),!,
	Plain=..[Functor|Args].
term_to_elements(Term, M, Functor, Args):-
	atomic(Functor), nonvar(Args),
	Term1 =.. [Functor|Args],
	Term = M:Term1.


%%	find_declaring_module(+ContextModule,+TermHead,-DeclarindModule)
%
%	Finds module name that declares TermHead predicate.
%
%	@param	ContextModule		Module into which TermHead predicate is imported.
%	@param	TermHead			Predicate head.
%	@param	DeclarindModule		TermHed is declared in DeclarindModule.
%
find_declaring_module(ContextModule,TermHead,ImportedFromModule):-
	ContextModule:(predicate_property(TermHead,
					  imported_from(ImportedFromModule))),!.
find_declaring_module(ContextModule,TermHead,ContextModule):-
	ContextModule:(current_predicate(_,TermHead)).


not(A):-A,!,false.
not(_).


:-op(900, fx, not).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


/* File:      odbc_call.P
** Author(s): Hasan Davulcu, Lily Dong, David S. Warren
** Contact:   xsb-contact@cs.sunysb.edu
**
** Copyright (C) The Research Foundation of SUNY, 1993-1998
**
** XSB is free software; you can redistribute it and/or modify it under the
** terms of the GNU Library General Public License as published by the Free
** Software Foundation; either version 2 of the License, or (at your option)
** any later version.
**
** XSB is distributed in the hope that it will be useful, but WITHOUT ANY
** WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
** FOR A PARTICULAR PURPOSE.  See the GNU Library General Public License for
** more details.
**
** You should have received a copy of the GNU Library General Public License
** along with XSB; if not, write to the Free Software Foundation,
** Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
**
** $Id: db_call.P,v 1.41 2007/01/24 22:15:42 dwarren Exp $
**
*/

:- dynamic attribute/5. %(_FldNo,_Pred,_Attr,_Type,_ConnHdl).
:- dynamic relation/4. %(_Pred,_PredArity,_Tab,_ConnHdl).
:- dynamic attributeX/4. %(_Tab,_Attr,_Type,_ConnHdl).
:- dynamic table_arity/3. %(_Conn,_Tab,_TabArity).
:- dynamic db_asserted_query/2. %(_ConnHand,_Head).

% --- for projection goal, we are to use bind variables for those constants.

get_mode([],[],[]) :- !.
get_mode([HA|TA],[bind(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	HA = 'NULL',!,
	writeln('WAR - DB: ''NULL'' as a bind variable.'),
	get_mode(TA,TM,TR).
get_mode([HA|TA],['NULL'(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	HA = 'NULL'(X),var(X),!,
	get_mode(TA,TM,TR).
get_mode([HA|TA],[bind(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	HA = 'NULL'(X),nonvar(X),!,
	get_mode(TA,TM,TR).
get_mode([HA|TA],[HR|TM],[HR|TR]) :-
	var(HA),!,
	get_mode(TA,TM,TR).
get_mode([HA|TA],[bind(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	get_mode(TA,TM,TR).

% --- basically same as predicate get_mode.  the special case we take care of here is that if
% --- the SQL query is an insertion, then we don't distinguish 'NULL'(Var) or 'NULL'(constant)

get_mode_ins([],[],[]) :- !.
get_mode_ins([HA|TA],[bind(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	HA = 'NULL',!,
	writeln('WAR - DB: ''NULL'' as a bind variable.'),
	get_mode_ins(TA,TM,TR).
get_mode_ins([HA|TA],['NULL'(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	HA = 'NULL'(_X),!,
	get_mode_ins(TA,TM,TR).
get_mode_ins([HA|TA],[HR|TM],[HR|TR]) :-
	var(HA),!,
	get_mode_ins(TA,TM,TR).
get_mode_ins([HA|TA],[bind(HR)|TM],[HR|TR]) :-
	nonvar(HA),
	get_mode_ins(TA,TM,TR).

% --------------------------------------------------------------------------------------
%
% This Prolog to SQL compiler may be distributed free of charge provided that it is
% not used in commercial applications without written consent of the author, and
% that the copyright notice remains unchanged.
%
%                    (C) Copyright by Christoph Draxler, Munich
%                        Version 1.1 of Dec. 21st 1992
%
% I would like to keep in my hands the further development and distribution of the
% compiler. This does not mean that I don't want other people to suggest or even
% implement improvements - quite on the contrary: I greatly appreciate contributions
% and if they make sense to me I will incorporate them into the compiler (with due
% credits given!).
%
% For further development of the compiler, address your requests, comments and
% criticism to the author:
%
%                    Christoph Draxler
%                    CIS Centre for Information and Speech Processing
%                    Ludwig-Maximilians-University Munich
%                    Wagmuellerstr. 23
%                    D 80538 Munich
%                    Tel : ++49 / +89 / 211 06 64 (-60)
%                    Fax : ++49 / +89 / 211 06 74
%                    Mail: draxler@cis.uni-muenchen.de
%
%
% A report describing the implementation is available upon request from the
% author.
%
%
% RELEASE INFORMATION
% ===================
% Current version is v. 1.1 of Dec. 21st 1992.
% Version 1.0 Sept. 3 1992
% --------------------------------------------------------------------------------------

% --- banchmarks of sample queries ---------
% --- Meta Database for schema definition of SQL DB in Prolog --------------------------
%
% maps Prolog predicates to SQL table names, Prolog predicate argument positions to SQL
% attributes, and Prolog operators to SQL operators.
%
% ATTENTION! It is assumed that the arithmetic operators in Prolog and SQL are the same,
% i.e. + is addition in Prolog and in SQL, etc. If this is not the case, then a mapping
% function for arithmetic operators is necessary too.
% ---------------------------------------------------------------------off -----------------

%%	add_to_dictionary(Key,RangeVar,Attribute,Quantifier,Dict,NewDict)

add_to_dictionary(Key,RangeVar,Attribute,Type,Quantifier,Dict,DictOut) :-
	(member(dict(Key,RangeVar1,Attribute1,Type1,Existential),Dict)
	 ->	RangeVar1 = RangeVar,
		Attribute1 = Attribute,
		Type1 = Type,
		Existential = existential,
		DictOut = Dict
	 ;	DictOut = [dict(Key,RangeVar,Attribute,Type,Quantifier)|Dict]
	).

%%	lookup(Key,Dict,Value)

lookup(VarId,Dict,RangeVar,Attribute,Type) :-
	member(dict(VarId,RangeVar,Attribute,Type,Quant),Dict),
	(Quant = all
	 ->	true
	 ;	nonvar(RangeVar),
		nonvar(Attribute)
	).

% --- set_difference(SetA,SetB,Difference) --------------------------------------------
%
% SetA - SetB = Difference
%--------------------------------------------------------------------------------------

set_difference([],_,[]).
set_difference([X|S1],S2,S3) :-
	(member(X,S2)
	 ->	S3 = S3a
	 ;	S3 = [X|S3a]
	),
	set_difference(S1,S2,S3a).


% --------------------------------------------------------------------------------------
%
% Output to screen predicates - rather crude at the moment
%
% --------------------------------------------------------------------------------------

% --- printqueries(Code) ---------------------------------------------------------------

printqueries([Query]) :-
	print_query(Query),
	write(';'),
	nl.
printqueries([Query|Queries]) :-
	not (Queries = []),
	print_query(Query),
	write('UNION'),
	printqueries(Queries).

%%	print_query(QueryCode)

print_query(query([agg_query(Function,Select,From,Where,Group)],_,_)) :-
	%% --- ugly rule here: aggregate function only in SELECT Part of query ----
	!, print_query(agg_query(Function,Select,From,Where,Group)).
print_query(query(Select,From,Where)) :-
	print_clause('SELECT',Select,','),
	print_clause('FROM',From,','),
	print_clause('WHERE',Where,'AND').
print_query(del_query(From,Where)) :-
	print_clause('DELETE FROM',From,','),
	print_clause('WHERE',Where,'AND').
print_query(ins_query(From,Columns,Where)):-
	print_ins_clause('INSERT INTO',From,','),
	print_ins_clause('(',Columns,','),
	write(')'),
	print_ins_clause('VALUES (',Where,','),
	write(')').
print_query(agg_query(Function,Select,From,Where,Group)) :-
	print_clause('SELECT',Function,Select,','),
	print_clause('FROM',From,','),
	print_clause('WHERE',Where,'AND'),
	print_clause('GROUP BY',Group,',').
print_query(negated_existential_subquery(Select,From,Where)) :-
	write('NOT EXISTS'),
	write('('),
	print_clause('SELECT',Select,','),
	print_clause('FROM',From,','),
	print_clause('WHERE',Where,'AND'),
	write(')').

%%	print_clause(Keyword,ClauseCode,Separator)
%
%	with
%	Keyword    one of SELECT, FROM, WHERE, or GROUP BY,
%	ClauseCode the code corresponding to the appropriate clause of an SQL query,
%	and Separator indicating the character(s) through which the items of
%	a clause are separated from each other (',' or 'AND').
%
print_clause(_Keyword,[],_).
print_clause(Keyword,[Column|RestColumns],Separator) :-
	write(' '),
	write(Keyword),
	write(' '),
	print_clause([Column|RestColumns],Separator).
print_clause(Keyword,Function,[Column],Separator) :-
	write(' '),
	write(Keyword),
	write(' '),
	write(Function),
	write('('),
	print_clause([Column],Separator),
	write(')').

%%	print_clause(ClauseCode,Separator)

print_clause([Item],_) :-
	print_column(Item).
print_clause([Item,NextItem|RestItems],Separator) :-
	print_column(Item),
	write(' '),
	write(Separator),
	write(' '),
	print_clause([NextItem|RestItems],Separator).

print_ins_clause(_Keyword,[],_).
print_ins_clause(Keyword,[Column|RestColumns],Separator) :-
	write(Keyword),
	write(' '),
	print_ins_clause([Column|RestColumns],Separator).

print_ins_clause([Item],_) :-
	print_ins_column(Item).
print_ins_clause([Item,NextItem|RestItems],Separator) :-
	print_ins_column(Item),
	write(' '),
	write(Separator),
	write(' '),
	print_ins_clause([NextItem|RestItems],Separator).

% --- print_column(ColumnCode) --------------------------------

print_column('*') :-
	write('*').
print_column(att(RangeVar,Attribute, _Type)) :-
	write(RangeVar),
	write('.'),
	write(Attribute).
print_column(rel(Relation,RangeVar)) :-
	write(Relation),
	write(' '),
	write(RangeVar).
print_column('$const$'('NULL')) :-
	!,
	write('NULL').
print_column('$const$'(String)) :-
	get_type('$const$'(String),string),
	write(''''),
	write(String),
	write('''').
print_column('$const$'(Number)) :-
	get_type('$const$'(Number),NumType),
	type_compatible(NumType,number),
	write(Number).
print_column('$bind$'(_BindId, _Type)) :-
	write('?').
print_column(comp(LeftArg,Operator,RightArg)) :-
	print_column(LeftArg),
	write(' '),
	write(Operator),
	write(' '),
	print_column(RightArg).
print_column(LeftExpr * RightExpr) :-
	print_column(LeftExpr),
	write('*'),
	print_column(RightExpr).
print_column(LeftExpr / RightExpr) :-
	print_column(LeftExpr),
	write('/'),
	print_column(RightExpr).
print_column(LeftExpr + RightExpr) :-
	print_column(LeftExpr),
	write('+'),
	print_column(RightExpr).
print_column(LeftExpr - RightExpr) :-
	print_column(LeftExpr),
	write('-'),
	print_column(RightExpr).
print_column(agg_query(Function,Select,From,Where,Group)) :-
	write('('),
	print_query(agg_query(Function,Select,From,Where,Group)),
	write(')').
print_column(negated_existential_subquery(Select,From,Where)) :-
	print_query(negated_existential_subquery(Select,From,Where)).

print_ins_column(rel(Relation,_RangeVar)) :-
	write(Relation).
print_ins_column(comp(_LeftArg,_Operator,RightArg)) :-
	print_column(RightArg).

% --- Mapping of Prolog operators to SQL operators -------------------------------------

comparison(=,=).
comparison(<,<).
comparison(>,>).
comparison(@<,<).
comparison(@>,>).
comparison('IS','IS').

negated_comparison(=,'<>').
negated_comparison(\=,=).
negated_comparison(>,=<).
negated_comparison(=<,>).
negated_comparison(<,>=).
negated_comparison(>=,<).

% --- aggregate_function(PrologFunctor,SQLFunction) -----------------

aggregate_functor(avg,'AVG').
aggregate_functor(min,'MIN').
aggregate_functor(max,'MAX').
aggregate_functor(sum,'SUM').
aggregate_functor(count,'COUNT').

% --- type system --------------------------------------------------------------
%
% A rudimentary type system is provided for consistency checking during the
% translation and for output formatting
%
% The basic types are string and number. number has the subtypes integer and
% real.
%
% ------------------------------------------------------------------------------

type_compatible(Type,Type) :-
	is_type(Type).
type_compatible(SubType,Type) :-
	subtype(SubType,Type).
type_compatible(Type,SubType) :-
	subtype(SubType,Type).

%%	subtype(SubType,SuperType)
%
% Simple type hierarchy checking
%
subtype(SubType,Type) :-
	is_subtype(SubType,Type).
subtype(SubType,Type) :-
	is_subtype(SubType,InterType),
	subtype(InterType,Type).

% --- is_type(Type) ------------------------------------------------------------
%
% Type names
%
% ------------------------------------------------------------------------------

is_type(number).
is_type(integer).
is_type(real).
is_type(string).
is_type(natural).

% --- is_subtype(SubType,Type) --------------------------------------------
%
% Simple type hierarchy for numeric types
%
% ------------------------------------------------------------------------------

is_subtype(integer,number).
is_subtype(real,number).
is_subtype(natural,integer).

% --- get_type(Constant,Type) --------------------------------------------------
%
% Prolog implementation specific definition of type retrieval
% sepia Prolog version given here
%
% ------------------------------------------------------------------------------

get_type(Type,TypeOut) :-
	atom(Type),
	!,
	atom_codes(Type,TypeCodes),
	(append("int",_,TypeCodes)
	 ->	TypeOut = integer
	 ; append("INT",_,TypeCodes)
	 ->	TypeOut = integer
	 ; append("num",_,TypeCodes)
	 ->	TypeOut = number
	 ; append("NUM",_,TypeCodes)
	 ->	TypeOut = number
	 ;	TypeOut = string
	).
get_type('$const$'(Constant),TypeOut) :-
	(integer(Constant)
	 ->	TypeOut = integer
	 ; number(Constant)
	 ->	TypeOut = number
	 ; atom(Constant)
	 ->	TypeOut = string
	).

% --- gensym(Root,Symbol) ----------------------------------------------------
%
% (C) Christoph Draxler, Aug. 1992
%
% ----------------------------------------------------------------------------

init_gensym(Root) :-
	nonvar(Root),
	nb_setval(Root,0).

gensym(Root,Symbol) :-
	nonvar(Root),
	var(Symbol),
	nb_getval(Root,Counter),
	NewCounter is Counter + 1,
	nb_setval(Root,NewCounter),
	number_codes(NewCounter,NCCodes),
	atom_codes(Root,RootCodes),
	append(RootCodes,NCCodes,NewCodes),
	atom_codes(Symbol,NewCodes).

% --- tokenize_term(Term,TokenizedTerm) -------------------------------------------------
%
% If Term is a
%
%  - variable, then this variable is instantiated with a unique identifier
%    of the form '$var$'(VarId), and TokenizedTerm is bound to the same
%    term '$var$'(VarId).
%
%  - constant, then TokenizedTerm is bound to '$const$'(Term).
%
%  - complex term, then the term is decomposed, its arguments are tokenized,
%    and TokenizedTerm is bound to the result of the composition of the original
%    functor and the tokenized arguments.
%
% --------------------------------------------------------------------------------------
%

tokenize_term(M:Term,M:TokenizedTerm):-!,
	tokenize_term1(M, M:Term, M:TokenizedTerm).


tokenize_term1(_,'$var$'(VarId),'$var$'(VarId)) :- !,
	(var(VarId)
	 ->	gensym(var,VarId)
	 ;	true
	).
tokenize_term1(_,bind(BindId),'$bind$'(BindId,_Type)) :- !,
	(var(BindId)
	 ->	gensym('BIND', BindId)
	 ;	true
	).
tokenize_term1(_,'$bind$'(BindId),'$bind$'(BindId,_Type)) :- !,
	nonvar(BindId).
tokenize_term1(_,'NULL'(HR),'$const$'('NULL')) :- !,
	HR = '$const$'('NULL').
tokenize_term1(_,'$const$'('NULL'),'$const$'('NULL')) :- !.
tokenize_term1(_,A,'$bind$'(A,_Type)) :-
	atom(A),
	atom_codes(A,[66,73,78,68|_]),   %% BIND_
	!.
tokenize_term1(_,Constant,'$const$'(Constant)) :-
	atomic(Constant),
	!.

tokenize_term1(ContextModule,M:Term,M:TokenizedTerm) :-!,
	nonvar(Term),
	Term =.. [Functor|Arguments],
	tokenize_arguments(ContextModule,Arguments,TokenArguments),
	TokenizedTerm =.. [Functor|TokenArguments].

tokenize_term1(ContextModule,Term,TokenizedTerm):-
	functor(Term,F,A),functor(T1,F,A),
	find_declaring_module(ContextModule,T1,M),
	tokenize_term1(ContextModule,M:Term,TokenizedTerm).

% --- tokenize_arguments(Arguments,TokenizedArguments) ---------------------------------
%
% organizes tokenization of arguments by traversing list and calling tokenize_term
% for each element of the list.
%
% --------------------------------------------------------------------------------------

tokenize_arguments(_,[],[]).
tokenize_arguments(ContextModule,[FirstArg|RestArgs],[TokFirstArg|TokRestArgs]) :-
	tokenize_term1(ContextModule,FirstArg,TokFirstArg),
	tokenize_arguments(ContextModule,RestArgs,TokRestArgs).

% --------------------------------------------------------------------------------------
%
% Top level predicate translate/5 organizes the compilation and constructs a
% Prolog term representation of the SQL query.
%
% --------------------------------------------------------------------------------------

translate(ProjectionTerm,DatabaseGoal,SQLQueryTerm,SQLAtom,Blist) :-
	%% --- initialize variable identifiers and range variables for relations -----
	init_gensym(var),
	init_gensym(rel),
	init_gensym('BIND'),
	%% --- tokenize projection term and database goal ----------------------------
	tokenize_term(ProjectionTerm,TokenProjectionTerm),
	tokenize_term(DatabaseGoal,TokenDatabaseGoal),
	%% --- lexical analysis: reordering of goals for disjunctive normalized form -
	disjunction(TokenDatabaseGoal,Disjunction),
	%% --- code generation ----------------------------------------------
	query_generation(Disjunction,TokenProjectionTerm,SQLQueryTerm),
	(db_show_query(on) -> printqueries(SQLQueryTerm);true),
	queries_atom(SQLQueryTerm,SQLAtom,Blist).

%%	linearize(Goal,ConjunctionList)
%
% Returns a conjunction of base goals for a complex disjunctive or conjunctive goal
% Yields several solutions upon backtracking for disjunctive goals
%

%  transform left-linear to right-linear conjunction (',' is associative)
linearize(((A,B),C),(LinA,(LinB,LinC))) :- !,
	linearize(A,LinA),
	linearize(B,LinB),
	linearize(C,LinC).
linearize((A,B),(LinA,LinB)) :- !,
	%% make sure A is not a conjunction
	linearize(A,LinA),
	linearize(B,LinB).
linearize((A;B),LinAB) :- !,
	linearize(A,LinA),
	linearize(B,LinB),
	(LinAB = LinA
	 ;
	 LinAB = LinB
	).
linearize(not A, not LinA) :- !,
	linearize(A,LinA).
linearize(Var^A, Var^LinA) :- !,
	linearize(A,LinA).
linearize(A,A).

% --- disjunction(Goal,Disjunction) -----------------------------------------------------
%
% turns original goal into disjunctive normalized form by computing all conjunctions
% and collecting them in a list
%
% ---------------------------------------------------------------------------------------

disjunction(Goal,Disjunction) :-
	findall(Conjunction,linearize(Goal,Conjunction),Disjunction).

% --- query_generation(ListOfConjunctions, ProjectionTerm, ListOfQueries) --------------
%
% For each Conjunction translate the pair (ProjectionTerm,Conjunction) to an SQL query
% and connect each such query through a UNION-operator to result in the ListOfQueries.
%
% A Conjunction consists of positive or negative subgoals. Each subgoal is translated
% as follows:
%  - the functor of a goal that is not a comparison operation is translated to
%    a relation name with a range variable
%  - negated goals are translated to NOT EXISTS-subqueries with * projection
%  - comparison operations are translated to comparison operations in the WHERE-clause
%  - aggregate function terms are translated to aggregate function (sub)queries
%
% The arguments of a goal are translated as follows:
%  - variables of a goal are translated to qualified attributes
%  - variables occurring in several goals are translated to equality comparisons
%    (equi join) in the WHERE-clause
%  - constant arguments are translated to equality comparisons in the WHERE-clause
%
% Special treatment of arithmetic functions:
%  - arithmetic functions are identified through the Prolog is/2 operator
%  - an arithmetic function may contain an unbound variable only on its left side
%  - the right side of the is/2 operator may consist of
%    * bound variables (bound through occurrence within a positive database goal, or
%      bound through preceeding arithmetic function), or of
%    * constants (numbers, i.e. integers, reals)
%
% The following RESTRICTION holds:
%
%  - the binding of variables follows Prolog: variables are bound by positive base goals
%    and on the left side of the is/2 predicate - comparison operations, negated goals
%    and right sides of the is/2 predicate do not return variable bindings and may even
%    require all arguments to be bound for a safe evaluation.
%
% --------------------------------------------------------------------------------------

query_generation([],_,[]).
query_generation([Conjunction|Conjunctions],ProjectionTerm,[Query|Queries]) :-
	projection_term_variables(ProjectionTerm,InitDict),
	translate_conjunction(Conjunction,SQLFrom,SQLWhere,InitDict,Dict),
	translate_projection(ProjectionTerm,Dict,SQLSelect),
	Query = query(SQLSelect,SQLFrom,SQLWhere),
	query_generation(Conjunctions,ProjectionTerm,Queries).

% --- projection_term_variables(ProjectionTerm,Dict) -----------------------------------
%
% extracts all variables from the ProjectionTerm and places them into the
% Dict as a dict/4 term with their Identifier, a non instantiated RangeVar and
% Attribute argument, and the keyword existential for the type of quantification
%
% --------------------------------------------------------------------------------------

projection_list_vars([],[]).
projection_list_vars(['$var$'(VarId)|RestArgs],[dict(VarId,_,_,_,existential)|RestVars]) :-
	projection_list_vars(RestArgs,RestVars).
projection_list_vars(['$const$'(_)|RestArgs],Vars) :-
	projection_list_vars(RestArgs,Vars).
projection_list_vars(['$bind$'(BindId,Type)|RestArgs],[dict(BindId,_,_, Type,existential)|RestVars]) :-
	projection_list_vars(RestArgs,RestVars).

projection_term_variables('$const$'(_),[]) :- !.
projection_term_variables('$var$'(VarId),[dict(VarId,_,_,_,existential)]) :- !.
projection_term_variables('$bind$'(BindId,Type),[dict(BindId,_,_,Type,existential)]) :- !.
projection_term_variables(_M:ProjectionTerm,ProjectionTermVariables) :-
	ProjectionTerm =.. [_Functor|ProjectionTermList],
	projection_list_vars(ProjectionTermList,ProjectionTermVariables).

% --- translate_conjunction(Conjunction,SQLFrom,SQLWhere,Dict,NewDict) -----------------
%
% translates a conjunction of goals (represented as a list of goals preceeded by
% existentially quantified variables) to FROM- and WHERE-clause of an SQL query.
% A dictionary containing the associated SQL table and attribute names is built up
% as an accumulator pair (arguments Dict and NewDict)
%
% --------------------------------------------------------------------------------------

translate_conjunction('$var$'(VarId)^Goal,SQLFrom,SQLWhere,Dict,NewDict) :-
	!,
	%% --- add info on existentially quantified variables to dictionary here -------------
	add_to_dictionary(VarId,_,_,_,existential,Dict,TmpDict),
	translate_conjunction(Goal,SQLFrom,SQLWhere,TmpDict,NewDict).
translate_conjunction(_M:(Goal,Conjunction),SQLFrom,SQLWhere,Dict,NewDict) :-
	!,
	translate_goal(Goal,FromBegin,WhereBegin,Dict,TmpDict),
	translate_conjunction(Conjunction,FromRest,WhereRest,TmpDict,NewDict),
	append(FromBegin,FromRest,SQLFrom),
	append(WhereBegin,WhereRest,SQLWhere).
translate_conjunction(Goal,SQLFrom,SQLWhere,Dict,NewDict) :-
	translate_goal(Goal,SQLFrom,SQLWhere,Dict,NewDict).

% --- translate_goal(Goal,SQLFrom,SQLWhere,Dict,NewDict) -------------------------------
%
% translates a
%
%   - positive database goal to the associated FROM- and WHERE clause of an SQL query
%   - a negated goal to a negated existential subquery
%   - an arithmetic goal to an arithmetic expression or an aggregate function query
%   - a comparison goal to a comparison expression
%   - a negated comparison goal to a comparison expression with the opposite comparison
%     operator
%
% --------------------------------------------------------------------------------------

translate_goal(M:SimpleGoal,[SQLFrom],SQLWhere,Dict,NewDict) :-
	% --- positive goal binds variables - these bindings are held in the dictionary -----
	functor(SimpleGoal,Functor,Arity),
	translate_functor(M:Functor,Arity,SQLFrom),
	SimpleGoal =.. [Functor|Arguments],
	translate_arguments(M:Functor,Arguments,SQLFrom,1,SQLWhere,Dict,NewDict).
translate_goal(_M:(Result is Expression),[],SQLWhere,Dict,NewDict) :-
	translate_arithmetic_function(Result,Expression,SQLWhere,Dict,NewDict).
translate_goal(_M:(not M:NegatedGoals),[],SQLNegatedSubquery,Dict,Dict) :-
	% --- negated goals do not bind variables - hence Dict is returned unchanged--------
	functor(NegatedGoals,Functor,_),
	not comparison(Functor,_),
	translate_conjunction(M:NegatedGoals,SQLFrom,SQLWhere,Dict,_),
	SQLNegatedSubquery = [negated_existential_subquery([*],SQLFrom,SQLWhere)].
translate_goal(_M:(not _M1:ComparisonGoal),[],SQLCompOp,Dict,Dict) :-
	% --- comparison operations do not bind variables - Dict is returned unchanged ------
	ComparisonGoal =.. [ComparisonOperator,LeftArg,RightArg],
	comparison(ComparisonOperator,SQLOperator),
	negated_comparison(SQLOperator,SQLNegOperator),
	translate_comparison(LeftArg,RightArg,SQLNegOperator,Dict,SQLCompOp).
translate_goal(_M:ComparisonGoal,[],SQLCompOp,Dict,Dict) :-
	% --- comparison operations do not bind variables - Dict is returned unchanged ------
	ComparisonGoal =.. [ComparisonOperator,LeftArg,RightArg],
	comparison(ComparisonOperator,SQLOperator),
	translate_comparison(LeftArg,RightArg,SQLOperator,Dict,SQLCompOp).

% --- translate_functor(Functor,QualifiedTableName) ------------------------------------
%
% translate_functor searches for the matching relation table name for
% a given functor and creates a unique range variable to result in
% a unique qualified relation table name.
%
% ----------------------------------------------------------------------

translate_functor(M:Functor,Arity,rel(TableName,RangeVariable)) :-
	relation(M:Functor,Arity,TableName,_),
	gensym(rel,RangeVariable).

%%	translate_arguments(Arguments,RelTable,ArgPos,Conditions,Dict)
%
% translate_arguments organizes the translation of term arguments. One
% term argument after the other is taken from the list of term arguments
% until the list is exhausted.
%
translate_arguments(_,[],_,_,[],Dict,Dict).
translate_arguments(Functor,[Arg|Args],SQLTable,Position,SQLWhere,Dict,NewDict) :-
	translate_argument(Functor,Arg,SQLTable,Position,Where,Dict,TmpDict),
	NewPosition is Position + 1,
	translate_arguments(Functor,Args,SQLTable,NewPosition,RestWhere,TmpDict,NewDict),
	append(Where,RestWhere,SQLWhere).

%%	translate_argument(Argument,RelTable,Position,Condition,Dict)
%
% The first occurrence of a variable leads to its associated SQL attribute information
% to be recorded in the Dict. Any further occurrence creates an equi-join condition
% between the current attribute and the previously recorded attribute.
% Constant arguments always translate to equality comparisons between an attribute and
% the constant value.
%
translate_argument(Functor,'$bind$'(BindId,Type),rel(_SQLTable,RangeVar),Position,Bind_Cond,Dict,NewDict) :-
	attribute(Position,Functor,Attribute,Type,_),
	add_to_dictionary(BindId,RangeVar,Attribute,Type,all,Dict,NewDict),
	Bind_Cond = [comp(att(RangeVar,Attribute,Type),=,'$bind$'(BindId,Type))].
translate_argument(Functor,'$bind$'(BindId,Type),rel(_SQLTable,RangeVar),Position,AttComparison,Dict,Dict) :-
	%% Variable occurred previously:
	%%	      retrieve first occurrence data from dictionary
	lookup(BindId,Dict,PrevRangeVar,PrevAtt,PrevType),
	attribute(Position,Functor,Attribute,Type,_),
	type_compatible(PrevType,Type),
	AttComparison = [comp(att(RangeVar,Attribute,Type),=,att(PrevRangeVar,PrevAtt,PrevType))].
translate_argument(Functor,'$var$'(VarId),rel(_SQLTable,RangeVar),Position,[],Dict,NewDict) :-
	attribute(Position,Functor,Attribute,Type,_),
	add_to_dictionary(VarId,RangeVar,Attribute,Type,all,Dict,NewDict).
translate_argument(Functor,'$var$'(VarId),rel(_SQLTable,RangeVar),Position,AttComparison,Dict,Dict) :-
	%% Variable occurred previously:
	%%	      retrieve first occurrence from data dictionary
	lookup(VarId,Dict,PrevRangeVar,PrevAtt,PrevType),
	attribute(Position,Functor,Attribute,Type,_),
	type_compatible(PrevType,Type),
	AttComparison = [comp(att(RangeVar,Attribute,Type),=,att(PrevRangeVar,PrevAtt,PrevType))].
translate_argument(Functor,'$const$'('NULL'),rel(_SQLTable,RangeVar),Position,ConstComparison,Dict,Dict) :-
	!,
	%% -- Equality comparison of constant value and attribute in table
	attribute(Position,Functor,Attribute,Type,_),
	get_type('$const$'('NULL'),_ConstType),	% do we need it here?
	ConstComparison = [comp(att(RangeVar,Attribute,Type),'IS','$const$'('NULL'))].
translate_argument(Functor,'$const$'(Constant),rel(_SQLTable,RangeVar),Position,ConstComparison,Dict,Dict) :-
	%% -- Equality comparison of constant value and attribute in table
	attribute(Position,Functor,Attribute,Type,_),
	get_type('$const$'(Constant),ConstType),
	type_compatible(ConstType,Type),
	ConstComparison = [comp(att(RangeVar,Attribute,Type),=,'$const$'(Constant))].

%%	translate_arithmetic_function(Result,Expression,SQLWhere,Dict,NewDict)
%
% Arithmetic functions (left side of is/2 operator is bound to value of expression on
% right side) may be called with either
%
% - Result unbound: then Result is bound to the value of the evaluation of Expression
% - Result bound: then an equality condition is returned between the value of Result
%   and the value of the evaluation of Expression.
%
% Only the equality test shows up in the WHERE clause of an SQLquery.
%
translate_arithmetic_function('$var$'(VarId),Expression,[],Dict,NewDict) :-
	%% assigment of value of arithmetic expression to variable - does not
	%% show up in WHERE-part, but expression corresponding to
	%% variable must be stored in Dict for projection translation
	evaluable_expression(Expression,Dict,ArithExpression,Type),
	add_to_dictionary(VarId,is,ArithExpression,Type,all,Dict,NewDict).
translate_arithmetic_function('$var$'(VarId),Expression,ArithComparison,Dict,Dict) :-
	%% Test if left side evaluates to right side: return equality comparison
	%% Left side consists of qualified attribute, i.e. range variable
	%% must not be arithmetic operator is/2
	lookup(VarId,Dict,PrevRangeVar,PrevAtt,PrevType),
	not (PrevRangeVar = is),
	%% test whether type of attribute is numeric - if not, there's no sense in
	%% continuing the translation
	type_compatible(PrevType,number),
	evaluable_expression(Expression,Dict,ArithExpression,ExprType),
	type_compatible(ExprType,number),
	ArithComparison = [comp(att(PrevRangeVar,PrevAtt, _),'=',ArithExpression)].
translate_arithmetic_function('$var$'(VarId),Expression,ArithComparison,Dict,Dict) :-
	%% test if left side evaluates to right side: return equality comparison
	%% Left side consists of arithmetic expression, i.e. VarId is stored
	%% in Dict as belonging to arithmetic expression which is expressed as
	%% RangeVar-argument of lookup returning is/2.
	%% Type information is implicit through the is/2 functor
	lookup(VarId,Dict,is,LeftExpr,Type),
	type_compatible(Type,number),
	evaluable_expression(Expression,Dict,RightExpr,ExprType),
	type_compatible(ExprType,number),
	ArithComparison = [comp(LeftExpr,'=',RightExpr)].
translate_arithmetic_function('$const$'(Constant),Expression,ArithComparison,Dict,Dict) :-
	%% is/2 used to test whether left side evaluates to right side
	get_type('$const$'(Constant),ConstantType),
	type_compatible(ConstantType,number),
	evaluable_expression(Expression,Dict,ArithExpression,ExprType),
	type_compatible(ExprType,number),
	ArithComparison = [comp('$const$'(Constant),'=',ArithExpression)].
translate_arithmetic_function('$bind$'(BindId,Type),Expression,ArithComparison,Dict,Dict) :-
	%%  is/2 used to test whether left side evaluates to right side
	lookup(BindId,Dict,_,_,Type),
	type_compatible(Type,number),
	evaluable_expression(Expression,Dict,ArithExpression,ExprType),
	type_compatible(ExprType,number),
	ArithComparison = [comp('$bind$'(BindId,Type),'=',ArithExpression)].


%%	evaluable_expression(ExpressionTerm,Dictionary,Expression,Type)
%
%	evaluable_expression constructs SQL arithmetic expressions with
%	qualified attribute names from the Prolog arithmetic expression
%	term and the information stored in the dictionary.
%	The type of an evaluable function is returned in the argument Type.
%	The dictionary is not changed because it is used for lookup only.
%
evaluable_expression(AggregateFunctionTerm,Dictionary,AggregateFunctionExpression,number) :-
	aggregate_function(AggregateFunctionTerm,Dictionary,AggregateFunctionExpression).
evaluable_expression(LeftExp + RightExp,Dictionary,LeftAr + RightAr,number) :-
	evaluable_expression(LeftExp,Dictionary,LeftAr,number),
	evaluable_expression(RightExp,Dictionary,RightAr,number).
evaluable_expression(LeftExp - RightExp,Dictionary,LeftAr - RightAr,number) :-
	evaluable_expression(LeftExp,Dictionary,LeftAr,number),
	evaluable_expression(RightExp,Dictionary,RightAr,number).
evaluable_expression(LeftExp * RightExp,Dictionary,LeftAr * RightAr,number) :-
	evaluable_expression(LeftExp,Dictionary,LeftAr,number),
	evaluable_expression(RightExp,Dictionary,RightAr,number).
evaluable_expression(LeftExp / RightExp,Dictionary, LeftAr / RightAr,number) :-
	evaluable_expression(LeftExp,Dictionary,LeftAr,number),
	evaluable_expression(RightExp,Dictionary,RightAr,number).
evaluable_expression('$var$'(VarId),Dictionary,att(RangeVar,Attribute, _),Type) :-
	lookup(VarId,Dictionary,RangeVar,Attribute,Type),
	RangeVar \= is.
evaluable_expression('$var$'(VarId),Dictionary,ArithmeticExpression,Type) :-
	lookup(VarId,Dictionary,is,ArithmeticExpression,Type).
evaluable_expression('$const$'(Const),_,'$const$'(Const),ConstType) :-
	get_type('$const$'(Const),ConstType).
evaluable_expression('$bind$'(BindId,Type),_,'$bind$'(BindId,Type),number).

%%	translate_comparison(_LeftArg,RightArg,CompOp,Dict,SQLComparison)
%
%	translates the left and right arguments of a comparison term into the
%	appropriate comparison operation in SQL. The result type of each
%	argument expression is checked for type compatibility
%
translate_comparison(LeftArg,RightArg,CompOp,Dict,Comparison) :-
	evaluable_expression(LeftArg,Dict,LeftTerm,LeftArgType),
	evaluable_expression(RightArg,Dict,RightTerm,RightArgType),
	type_compatible(LeftArgType,RightArgType),
	Comparison = [comp(LeftTerm,CompOp,RightTerm)],
	(LeftArg = '$bind$'(_,RightArgType);
	LeftTerm = '$bind$'(_,RightArgType); true).

%% 	aggregate_function(AggregateFunctionTerm,Dict,AggregateFunctionQuery)
%
%	aggregate_function discerns five Prolog aggregate function terms:
%	count, avg, min, max, and sum.
%	Each such term is has two arguments: a variable indicating the attribute
%	over which the function is to be computed, and a goal argument
%	which must contain in at least one argument position the variable:
%
%	e.g.  avg(Seats,plane(Type,Seats))
%
%	These aggregate function terms correspond to the SQL built-in
%	aggregate functions.
%
%	RESTRICTION: AggregateGoal may only be conjunction of
%	(positive or negative) base goals
%
aggregate_function(AggregateFunctionTerm,Dict,AggregateFunctionExpression) :-
	AggregateFunctionTerm =.. [AggFunctor,AggVar,AggGoal],
	aggregate_functor(AggFunctor,SQLFunction),
	conjunction(AggGoal,AggConjunction),
	aggregate_query_generation(SQLFunction,AggVar,AggConjunction,Dict,AggregateFunctionExpression).

conjunction(Goal,Conjunction) :-
	disjunction(Goal,[Conjunction]).

%%	aggregate_query_generation(Function,FunctionVariable,AggGoal,Dict,AggregateQuery)
%
%	compiles the function variable (representing the attribute over which
%	the aggregate function is to be computed) and the aggregate goal
%	(representing the selection and join conditions for the computation
%	of the aggregate function) to an SQL aggregate function subquery.
%
%
aggregate_query_generation(count,'$const$'('*'),AggGoal,Dict,AggregateQuery) :-
	translate_conjunction(AggGoal,SQLFrom,SQLWhere,Dict,_TmpDict),
	%% ATTENTION! It is assumed that in count(*) aggregate query terms there cannot be
	%% free variables because '*' stands for "all arguments"
	AggregateQuery = agg_query(_Function,(count,['$const$'(*)]),SQLFrom,SQLWhere,[]).
aggregate_query_generation(Function,FunctionVariable,AggGoal,Dict,AggregateQuery) :-
	translate_conjunction(AggGoal,SQLFrom,SQLWhere,Dict,TmpDict),
	%% -- only vars occurring in aggregate goal are relevant to the translation
	%% of the function variable and the free variables in the goal.
	%% Thus subtract from TmpDict all entries of Dict
	set_difference(TmpDict,Dict,AggDict),
	translate_projection(FunctionVariable,AggDict,SQLSelect),
	translate_grouping(FunctionVariable,AggDict,SQLGroup),
	AggregateQuery = agg_query(Function,SQLSelect,SQLFrom,SQLWhere,SQLGroup).

%%	translate_grouping(FunctionVariable,Dict,SQLGroup)
%
% finds the free variables in the aggregate function term and collects their
% corresponding SQL qualified attributes in the SQLGroup list.
%
translate_grouping(FunctionVariable,Dict,SQLGroup) :-
	free_vars(FunctionVariable,Dict,FreeVariables),
	translate_free_vars(FreeVariables,SQLGroup).

%%	free_vars(FunctionVariable,Dict,FreeVarList)
%
%	A Variable is free if it neither occurs as the FunctionVariable,
%	nor is stored as existentially quantified (through ^/2 in the
%	original goal) in the dictionary
%
%	FreeVars contains for each variable the relevant attribute and
%	relation information contained in the dictionary
%
free_vars(FunctionVariable,Dict,FreeVarList) :-
	projection_term_variables(FunctionVariable,FunctionVariableList),
	findall((Var,Table,Attribute),
		(member(dict(Var,Table,Attribute,_Type,all),Dict),
		 not member(dict(Var,_,_,_,_),FunctionVariableList)),
		FreeVarList).

%%	translate_free_vars(FreeVars,SQLGroup)
%
% translates dictionary information on free variables to SQLGroup of aggregate
% function query
%
translate_free_vars([],[]).
translate_free_vars([(_VarId,Table,Attribute)|FreeVars],[att(Table,Attribute, _)|SQLGroups]) :-
	translate_free_vars(FreeVars,SQLGroups).

%%	translate_projection(Term,Dict,SelectList)
%
%	RESTRICTION! ProjectionTerm underlies the following restrictions:
%
%  * ProjectionTerm must have a functor other than the built-in
%    operators, i.e. ',',';', etc. are not allowed
%
%  * only variables and constants are allowed as arguments,
%    i.e. no structured terms
%
translate_projection('$var$'(VarId),Dict,SelectList) :- !,
	projection_arguments(['$var$'(VarId)],SelectList,Dict).
translate_projection('$bind$'(BindId,Type),Dict,SelectList) :- !,
	projection_arguments(['$bind$'(BindId,Type)],SelectList,Dict).
translate_projection('$const$'(Const),_,['$const$'(Const)]) :- !.
translate_projection(_M:ProjectionTerm,Dict,SelectList) :-
	ProjectionTerm =.. [_Functor|Arguments],
	projection_arguments(Arguments,SelectList,Dict).

projection_arguments([],[],_).
projection_arguments([Arg|RestArgs],[Att|RestAtts],Dict) :-
	retrieve_argument(Arg,Att,Dict),
	projection_arguments(RestArgs,RestAtts,Dict).

%%	retrieve_argument(Argument,SQLAttribute,Dictionary)
%
%	retrieves the mapping of an argument to the appropriate SQL construct, i.e.
%
%  * qualified attribute names for variables in base goals
%  * arithmetic expressions for variables in arithmetic goals
%  * constant values for constants
%
retrieve_argument('$var$'(VarId),Attribute,Dict) :-
	lookup(VarId,Dict,TableName,AttName,_),
	(TableName = is
	 ->	Attribute = AttName
	 ;	Attribute = att(TableName,AttName, _)
	).
retrieve_argument('$bind$'(BindId,Type),Attribute,Dict) :-
	lookup(BindId,Dict,TableName,AttName,Type),!,
	Attribute = att(TableName,AttName, _).
retrieve_argument('$bind$'(BindId,Type),'$bind$'(BindId,Type),Dict) :-
	lookup(BindId,Dict,TableName,AttName,Type),
	var(TableName), var(AttName).
retrieve_argument('$const$'(Constant),'$const$'(Constant),_).

%%	queries_atom(Queries,QueryAtom, Blist)
%
% queries_atom(Queries,QueryAtom,Blist) returns in its second argument
% the SQL query as a Prolog atom. For efficiency reasons, a list
% of ASCII codes is ceated as a difference list, and it is then
% transformed to an atom by name/2.  It also create a list of bind variables
% occurring in the QueryAtom so that we can bind them to the place holders '?'
% at a later time.
%
queries_atom(Queries,QueryList,Blist) :-
	queries_atom(Queries,QueryList,[],Blist,[]).

queries_atom([Query],QueryList,Diff,Blist,Olist) :- !,
	query_atom(Query,QueryList,Diff,Blist,Olist).
queries_atom([Query|Queries],QueryList,Diff,Blist,Olist) :-
	query_atom(Query,QueryList,X1,Blist,Tlist),
	X1 = [' UNION '|X2],
	queries_atom(Queries,X2,Diff,Tlist,Olist).

% --- query_atom(QueryCode) --------------------------------

query_atom(query([agg_query(Function,Select,From,Where,Group)],_,_),QueryList,Diff,Blist,Olist) :-
	%% --- ugly rule here: aggregate function only in SELECT Part of query ----
	!, query_atom(agg_query(Function,Select,From,Where,Group),QueryList,Diff,Blist,Olist).
query_atom(query(Select,From,Where),QueryList,Diff,Blist,Olist) :-
	clause_atom(' SELECT ',Select,',',QueryList,X1,Blist,Tl2),
	clause_atom(' FROM ',From,',',X1,X2,Tl2,Tl1),
	clause_atom(' WHERE ',Where,' AND ',X2,Diff,Tl1,Olist).
query_atom(query_upd(Update,From,Where),QueryList,Diff,Blist,Olist) :-
	clause_upd_atom(' UPDATE ',From,',',QueryList,X1,Blist,Tl2),
	clause_upd_atom(' SET ',Update,',',X1,X2,Tl2,Tl1),
	clause_upd_atom(' WHERE ',Where,' AND ',X2,Diff,Tl1,Olist).
query_atom(del_query(From,Where),QueryList,Diff,Blist,Olist) :-
	clause_atom(' DELETE FROM ',From,',',QueryList,X1,Blist,Tlist),
	clause_atom(' WHERE ',Where,' AND ',X1,Diff,Tlist,Olist).
query_atom(ins_query(From,Columns,Where),QueryList,Diff,Blist,Olist) :-
	clause_ins_atom(' INSERT INTO',From,',',QueryList,X1,Blist,Tlist),
	clause_ins_atom(' (',Columns,',',X1,X2,Tlist,Tlist1),
	clause_ins_atom(') VALUES (',Where,',',X2,X3,Tlist1,Olist),
	column_atom_a(')',X3,Diff,_,_).
query_atom(agg_query(Function,Select,From,Where,Group),QueryList,Diff,Blist,Olist) :-
	clause_atom(' SELECT ',Function,Select,',',QueryList,X1,Blist,Tl3),
	clause_atom(' FROM ',From,',',X1,X2,Tl3,Tl2),
	clause_atom(' WHERE ',Where,' AND',X2,X3,Tl2,Tl1),
	clause_atom(' GROUP BY ',Group,',',X3,Diff,Tl1,Olist).
query_atom(negated_existential_subquery(Select,From,Where),QueryList,Diff,Blist,Olist) :-
	column_atom_a(' NOT EXISTS(',QueryList,X1,_,_),
	clause_atom(' SELECT ',Select,',',X1,X2,Blist,Tl2),
	clause_atom(' FROM ',From,',',X2,X3,Tl2,Tl1),
	clause_atom(' WHERE ',Where,' AND ',X3,X4,Tl1,Olist),
	column_atom_a(')',X4,Diff,_,_).

%%	clause_atom(Keyword,ClauseCode,Junctor,CurrAtom,QueryAtom,Blist,Olist)
%
%
% @param Keyword	one of SELECT, FROM, WHERE, or GROUP BY,
% @param ClauseCode	the code corresponding to the appropriate clause of an SQL query, and
% @param Junctor	indicating the character(s) through which the items of a clause
%	are separated from each other (',' or 'AND').
% @param Olist		list of bind variables in QueryAtom
% @param Blist		new list of bind variables in CurrAtom
%
clause_atom(Keyword,Columns,Junctor,QueryList,Diff,Blist,Olist):-
	clause_atom1(reg,Keyword,Columns,Junctor,QueryList,Diff,Blist,Olist).

clause_atom1(_CA,_Keyword,[],_,QueryList,QueryList,Blist,Blist).
clause_atom1(CA,Keyword,[Column|RestColumns],Junctor,QueryList,Diff,Blist,Olist) :-
	column_atom(CA,Keyword,QueryList,X1,Blist,Tlist),
	column_atom_a(' ',X1,X2,_,_),
	clause_atom1(CA,[Column|RestColumns],Junctor,X2,X3,Tlist,Olist),
	column_atom_a(' ',X3,Diff,_,_).

clause_atom1(CA,Keyword,Function,[Column],Junctor,QueryList,Diff,Blist,Olist) :-
	column_atom(CA,Keyword,QueryList,X1,Blist,Tl2),
	column_atom_a(' ',X1,X2,_,_),
	column_atom(CA,Function,X2,X3,Tl2,Tl1),
	column_atom_a('(',X3,X4,_,_),
	clause_atom1(CA,[Column],Junctor,X4,X5,Tl1,Olist),
	column_atom_a(') ',X5,Diff,_,_).

clause_atom(Items,Junctor,QueryList,Diff,Blist,Olist) :-
	clause_atom1(column_atom,Items,Junctor,QueryList,Diff,Blist,Olist).
clause_atom1(CA,[Item],_,QueryList,Diff,Blist,Olist) :-
	column_atom(CA,Item,QueryList,Diff,Blist,Olist).
clause_atom1(CA,[Item,NextItem|RestItems],Junctor,QueryList,Diff,Blist,Olist) :-
	clause_atom1(CA,[NextItem|RestItems],Junctor,X4,Diff,Tl1,Olist),
	column_atom_a(' ',X3,X4,_,_),
	column_atom(CA,Junctor,X2,X3,Tl2,Tl1),
	column_atom_a(' ',X1,X2,_,_),
	column_atom(CA,Item,QueryList,X1,Blist,Tl2).

clause_ins_atom(_Keyword,[],_,QueryList,QueryList,Blist,Blist).
clause_ins_atom(Keyword,[Column|RestColumns],Junctor,QueryList,Diff,Blist,Olist) :-
	column_atom(reg,Keyword,QueryList,X1,Blist,Tlist),
	column_atom_a(' ',X1,X2,_,_),
	clause_ins_atom([Column|RestColumns],Junctor,X2,X3,Tlist,Olist),
	column_atom_a(' ',X3,Diff,_,_).
clause_ins_atom([Item],_,QueryList,Diff,Blist,Olist) :-
	column_ins_atom(Item,QueryList,Diff,Blist,Olist).
clause_ins_atom([Item,NextItem|RestItems],Junctor,QueryList,Diff,Blist,Olist) :-
	column_ins_atom(Item,QueryList,X1,Blist,Tl2),
	column_atom_a(' ',X1,X2,_,_),
	column_atom(reg,Junctor,X2,X3,Tl2,Tl1),
	column_atom_a(' ',X3,X4,_,_),
	clause_ins_atom([NextItem|RestItems],Junctor,X4,Diff,Tl1,Olist).

clause_upd_atom(Keyword,Columns,Junctor,QueryList,Diff,Blist,Olist):-
	clause_atom1(upd,Keyword,Columns,Junctor,QueryList,Diff,Blist,Olist).


column_atom(Kind,att(_RangeVar,Attribute, _),QueryList,Diff,Blist,Olist) :-
	nonvar(Kind),Kind = upd,!,
	column_atom(Kind,Attribute,QueryList,Diff,Blist,Olist).
column_atom(Kind,att(RangeVar,Attribute, _),QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,RangeVar,QueryList,X1,Blist,Olist),
	column_atom_a('.',X1,X2,_,_),
	column_atom(Kind,Attribute,X2,Diff,_Tlist,Olist).
column_atom(Kind,rel(Relation,RangeVar),QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,Relation,QueryList,X1,Blist,Tlist),
	column_atom_a(' ',X1,X2,_,_),
	column_atom(Kind,RangeVar,X2,Diff,Tlist,Olist).
column_atom(_Kind,'$const$'('NULL'),QueryList,Diff,Blist,Olist) :-
	get_type('$const$'('NULL'),string),!,
	column_atom_a('NULL',QueryList,Diff,Blist,Olist).
column_atom(Kind,'$const$'(String),QueryList,Diff,Blist,Olist) :-
	get_type('$const$'(String),string),
	column_atom_a('''',QueryList,X1,Blist,Tl2),
	column_atom(Kind,String,X1,X2,Tl2,Tl1),
	column_atom_a('''',X2,Diff,Tl1,Olist).
column_atom(Kind,'$const$'(Number),QueryList,Diff,Blist,Olist) :-
	get_type('$const$'(Number),NumType),
	type_compatible(NumType,number),
	column_atom(Kind,Number,QueryList,Diff,Blist,Olist).
column_atom(_Kind,'$bind$'(BindId,Type),QueryList,Diff,Blist,Olist) :-
	Blist = [BindId|Tlist],
	((Type = string) -> Atom = '''?'''; Atom = '?'),
	column_atom_a(Atom,QueryList,Diff,Tlist,Olist).
column_atom(Kind,comp(LeftArg,Operator,RightArg),QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,LeftArg,QueryList,X1,Blist,Tl2),
	column_atom_a(' ',X1,X2,_,_),
	column_atom(Kind,Operator,X2,X3,Tl2,Tl1),
	column_atom_a(' ',X3,X4,_,_),
	column_atom(Kind,RightArg,X4,Diff,Tl1,Olist).
column_atom(Kind,LeftExpr * RightExpr,QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,LeftExpr,QueryList,X1,Blist,Tlist),
	column_atom_a('*',X1,X2,_,_),
	column_atom(Kind,RightExpr,X2,Diff,Tlist,Olist).
column_atom(Kind,LeftExpr + RightExpr,QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,LeftExpr,QueryList,X1,Blist,Tlist),
	column_atom_a('+',X1,X2,_,_),
	column_atom(Kind,RightExpr,X2,Diff,Tlist,Olist).
column_atom(Kind,LeftExpr - RightExpr,QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,LeftExpr,QueryList,X1,Blist,Tlist),
	column_atom_a('-',X1,X2,_,_),
	column_atom(Kind,RightExpr,X2,Diff,Tlist,Olist).
column_atom(Kind,LeftExpr / RightExpr,QueryList,Diff,Blist,Olist) :-
	column_atom(Kind,LeftExpr,QueryList,X1,Blist,Tlist),
	column_atom_a('/',X1,X2,_,_),
	column_atom(Kind,RightExpr,X2,Diff,Tlist,Olist).
column_atom(_Kind,agg_query(Function,Select,From,Where,Group),QueryList,Diff,Blist,Olist) :-
	column_atom_a('(',QueryList,X1,_,_),
	query_atom(agg_query(Function,Select,From,Where,Group),X1,X2,Blist,Olist),
	column_atom_a(')',X2,Diff,_,_).
column_atom(_Kind,negated_existential_subquery(Select,From,Where),QueryList,Diff,Blist,Olist) :-
	query_atom(negated_existential_subquery(Select,From,Where),QueryList,Diff,Blist,Olist).
column_atom(_Kind,Atom,QueryList,Diff,Blist,Olist) :-
	column_atom_a(Atom,QueryList,Diff,Blist,Olist).


column_atom_a(Atom,[PAtom|Diff],Diff,Blist,Blist) :-
	(atom(Atom)
	 ->	PAtom = Atom
	 ; number(Atom)
	 ->	name(Atom,X1),
		atom_codes(PAtom,X1)
	).

column_ins_atom(rel(Relation,_RangeVar),QueryList,Diff,Blist,Olist) :-
	column_atom(reg,Relation,QueryList,Diff,Blist,Olist).
column_ins_atom(comp(_LeftArg,_Operator,RightArg),QueryList,Diff,Blist,Olist) :-
	column_atom(reg,RightArg,QueryList,Diff,Blist,Olist).

translate_del(DeleteTerm,DatabaseGoal,SQLQueryTerm,SQLAtom,Blist) :-
	%% -- initialize variable identifiers and range variables for relations
	init_gensym(var),
	init_gensym(rel),
	init_gensym('BIND'),
	%% -- tokenize projection term and database goal
	tokenize_term(DeleteTerm,TokenDeleteTerm),
	tokenize_term(DatabaseGoal,TokenDatabaseGoal),
	%% -- lexical analysis: reordering of goals for disjunctive normalized form
	disjunction(TokenDatabaseGoal,Disjunction),
	%% -- code generation ----
	query_generation_del(Disjunction,TokenDeleteTerm,SQLQueryTerm),
	(db_show_query(on) -> printqueries(SQLQueryTerm);true),
	queries_atom(SQLQueryTerm,SQLAtom,Blist).

query_generation_del([],_,[]).
query_generation_del([Conjunction|Conjunctions],ProjectionTerm,[Query|Queries]) :-
	projection_term_variables(ProjectionTerm,InitDict),
	translate_conjunction(Conjunction,SQLFrom,SQLWhere,InitDict,Dict),
	translate_projection(ProjectionTerm,Dict,_SQLSelect),
	Query = del_query(SQLFrom,SQLWhere),
	query_generation(Conjunctions,ProjectionTerm,Queries).

translate_ins(InsertTerm,DatabaseGoal,SQLQueryTerm,SQLAtom,Blist) :-
	%% -- initialize variable identifiers and range variables for relations
	init_gensym(var),
	init_gensym(rel),
	init_gensym('BIND'),
	%% -- tokenize projection term and database goal ---
	tokenize_term(InsertTerm,TokenInsertTerm),
	tokenize_term(DatabaseGoal,TokenDatabaseGoal),
	%% -- lexical analysis: reordering of goals for disjunctive normalized form
	disjunction(TokenDatabaseGoal,Disjunction),
	%% -- code generation ------------------------------------------------
	query_generation_ins(Disjunction,TokenInsertTerm,SQLQueryTerm),
	(db_show_query(on) -> printqueries(SQLQueryTerm);true),
	queries_atom(SQLQueryTerm,SQLAtom,Blist).

query_generation_ins([],_,[]).
query_generation_ins([M:Conjunction|Conjunctions],ProjectionTerm,[Query|Queries]) :-
	projection_term_variables(ProjectionTerm,InitDict),
	translate_conjunction(M:Conjunction,SQLFrom,SQLWhere,InitDict,Dict),
	translate_projection(ProjectionTerm,Dict,_SQLSelect),
	functor(Conjunction,Pred,_Arity),
	setof(rel(ColNum,ColName),
	      [Type,Con]^attribute(ColNum,M:Pred,ColName,Type,Con),
	      Cols),
	filterColName(Cols,ColNames),
	Query = ins_query(SQLFrom,ColNames,SQLWhere),
	query_generation(Conjunctions,ProjectionTerm,Queries).

filterColName([],[]).
filterColName([rel(_,N)|Xs],[rel(N,_)|Ys]) :-
	filterColName(Xs,Ys).

/************************************************************************/
/*									*/
/*	ODBC Interface Routines						*/
/*									*/
/************************************************************************/

:- dynamic db_error_flg/1. %(_).
:- dynamic db_show_query/1. %(_).
:- dynamic db_connection/2. %(_Name,_Handle).  %% stack of connections made
:- dynamic db_active_cursors/1. %(_Cursor). %% estimate of stack of active cursors

:-use_module(library(odbc)).


%%	db_open(DSN,UserName,Passwd)
%
%	Same ast db_open/4 with =|Connection = defaultConnection|=.
%
db_open(DSN,UserName,Passwd) :-
	db_open(DSN, UserName, Passwd, defaultConnection).


%%	db_open(+DSN,+UserName,+Passwd,+Connection)
%
%	Open ODBC connection to data source DSN.
%	Null value is set to 'NULL'(_),
%	see swi odbc library documentation for details.
%
%	@param	DSN			ODBC Data Source Name.
%	@param	UserName	Database user name.
%	@param	Passwd		User password.
%	@param	Connection	Connection alias -- must be unique
%	for all connections to any DSN.
%
%	Example usage is:
%	==
%	db_open('MyDatabase',someusername,
%			password,myData).
%	==
%
db_open(DSN,UserName,Passwd,Connection) :-
	odbc:odbc_connect(DSN, CH, [
			     user(UserName),
			     password(Passwd),
			     alias(Connection)
			    ]),
	odbc:odbc_set_connection(Connection, null('NULL'(_))),
	db_newconnection(Connection, CH).


%%	db_open(ConnectIn)
%
%	Same as db_open/2
%	with =|Connection = defaultConnection|=.
%
db_open(ConnectIn) :-
	db_open(ConnectIn,defaultConnection).


%%	db_open(ConnectIn, Connection)
%
%	Connect using connection string.
%	Null value is set to 'NULL'(_) see db_open/4.
%
%	@param	ConnectIn	Pass whole Connect string, e.g.
%	'DRIVER={Microsoft Access Driver (*.mdb)};DBQ=C:\\sutures.mdb;UID=;PWD=;'
%
%	@param	Connection	connection alias see also db_open/4.
%
db_open(ConnectIn, Connection) :-
	odbc:odbc_driver_connect(ConnectIn, CH, [alias(Connection)]),
	odbc:odbc_set_connection(Connection, null('NULL'(_))),
	db_newconnection(Connection, CH).


db_newconnection(Connection,ConnHandle) :-
	(db_connection(_,_)	% init if first connection
	 ->	true
	 ;	init_gensym(table1),init_gensym(null),
		(db_error_flg(_)
		 ->	true
		 ;	assert(db_error_flg(fail))
		),
		(db_show_query(_)
		 ->	true
		 ;	assert(db_show_query(on))
		)
	),
	(var(Connection)
	 ->	(\+ db_connection(ConnHandle, ConnHandle),
	         asserta(db_connection(ConnHandle,ConnHandle)))
	 ;	(db_connection(Connection, _)
		 ->	db_error('Connection handle is already in use.'),
			fail
		 ;	(\+ db_connection(Connection, ConnHandle),
		        asserta(db_connection(Connection,ConnHandle)))
	    )
	).


db_shutdown :-
	findall(Connection,db_connection(Connection,_),Connections),
	(member(Connection,Connections),
	 db_close(Connection),
	 fail
	 ;
	 true
	).

%%	db_close
%
%	Closes all opened connections.
db_close :-
	findall(Connection,
		db_connection(Connection,_ConnHandle),
		Connections),
	member(Connection,Connections),
	db_close(Connection),
	fail.
db_close.

%%	db_close(Connection)
%
%	Closes Connection see db_open/4.
db_close(Connection) :-
	db_connection_check(Connection,ConnHandle),
	!,
	%db_disconnect(ConnHandle),
	retractall(db_connection(Connection,ConnHandle)),
	retract_import_info(ConnHandle).

retract_import_info(ConnHandle) :-
	(relation(PredName,Arity,_TableName,ConnHandle),
	 abolish(PredName, Arity),
	 fail
	 ;
	 true
	),
	retractall(relation(_,_,_,ConnHandle)),
	(db_asserted_query(ConnHandle,Query),
	 strip_module(Query, M, Plain),
	 functor(Plain, QueryName, Arity),
	 abolish(M:QueryName, Arity),
	 fail
	 ;
	 true
	),
	retractall(db_asserted_query(ConnHandle,_)),
	retractall(table_arity(ConnHandle,_,_)),
	retractall(attribute(_,_,_,_,ConnHandle)),
	retractall(attributeX(_,_,_,ConnHandle)).

db_sql(BindVals,SQLStmt,Result) :-
	db_sql(defaultConnection,BindVals,SQLStmt,Result).

db_sql(Connection,BindVals,SQLStmt,Result) :-
	db_connection_check(Connection,ConnHandle),
	db_sql_i(ConnHandle,BindVals,SQLStmt,Result).

db_sql_cnt(Connection,BindVals,SQLStmt,Count) :-
	db_connection_check(Connection,ConnHandle),
	db_sql_ic(ConnHandle,BindVals,SQLStmt,Count).

db_connection_check(Connection,ConnHandle):-
	(db_connection(Connection,ConnHandle)
	 ->	true
	 ;	concat_atom(['Invalid ODBC connection: ',Connection],Msg),
		db_error(db_error('XXXXX',Msg))
	).

findall_db_sql(BindVals,SQLStmt,Result) :-
	findall_db_sql(defaultConnection,BindVals,SQLStmt,Result).

findall_db_sql(Connection,BindVals,SQLStmt,Result) :-
	db_connection_check(Connection,ConnHandle),
	findall_db_sql_i(ConnHandle,BindVals,SQLStmt,Result).

findall_db_sql_i(ConnHandle,BindVals,SQLStmt,Result):-
	findall(R,db_sql_i(ConnHandle,BindVals,SQLStmt,R),Results),
	member(Result,Results).


/* db_sql_i(+ConnHandle,+BindVals,+SQLStmt,?Result) */
/* The most general sql call: */
db_sql_i(ConnHandle,BindVals,SQLStmt,Result) :-
	sqlstmt_to_atom(SQLStmt, SqlToBind),
	bind(SqlToBind, BindVals, Sql),
	odbc:odbc_query(ConnHandle, Sql, Row),
	Row=..[_|Result].

db_sql_ic(ConnHandle,BindVals,SQLStmt,Result) :-
	db_sql_i(ConnHandle,BindVals,SQLStmt, [Result]).


%	db_sql_i supporting stuff
%
%
replace(Atom, OldSequence, NewSequences, Replaced, L) :-
	atomic_list_concat(L, OldSequence, Atom),

	insert2nd(L, NewSequences, RL),
	atomic_list_concat(RL, Replaced).

insert2nd(L1, Elem, L):-
	insert1(L1, Elem, [], L).

insert1([], _, L, L):-!.
insert1([Last], _, L, L2):-
	append(L, [Last], L2), !.
insert1([H|T], [Elem|Elems], L1, L) :-!,
	append(L1, [H, Elem], L2),
	insert1(T, Elems, L2, L).

bind(InSql, BindVals, Sql):-
	replace(InSql, '?', BindVals, Sql, _L).

sqlstmt_to_atom(SQLStmt, SQLStmt):-
	atom(SQLStmt), !.
sqlstmt_to_atom(SQLStmt, SQL) :-
	is_list(SQLStmt), atomic_list_concat(SQLStmt, SQL), !.

% just to notify that we do not handle all cases
sqlstmt_to_atom(_, _) :- throw('sqlstmt_to_atom: unknown case').


%---------------------------------------------------------------------------------
% DB Import (and attach)
%---------------------------------------------------------------------------------

:-meta_predicate db_import(?,:).
:-meta_predicate db_import(?,?,:,?).
:-meta_predicate db_import(?,?,:).

%%	db_import(+DbPredicate,:PredName)
%
%	Same as db_import/3 with =|Connection = defaultConnection|=.
%
db_import(DbPredicate,PredName) :-
	db_import(defaultConnection,DbPredicate,PredName).


%%	db_import(+Connection,+DbPredicate,:PredName)
%
%	Same as db_import/4 with =|Options = [findall(false)]|=.
%
db_import(Connection,DbPredicate,PredName) :-
	db_import(Connection,DbPredicate,PredName,[findall(false)]).


%%	db_import(+Connection,+DbPredicate,:PredName,+Options)
%
%	Creates prolog predicate that represents 
%	given database table records.
%	See also XSB documentation.
%
%	Example for database table clients which have following columns
%	id, first_name, last_name. The prolog predicate =|client|=
%	that will represent clients table may be defined like this: 
%	==
%	db_import('clients'('id','first_name','last_name'),client).
%	==
%	Next the following query can be done,
%	which may backtrack through all
%	FirstName and LastName pairs from clients table:
%	==
%	client(_,FirstName,LastName).
%	==
%
%
%	@param	Connection	Connection alias see db_open/4.
%	@param	DbPredicate	Database table predicate.
%	@param	PredName	Prolog predicate name.
%	@param	Options		[findall(Boolean)] or [].
%
%
db_import(Connection,DbPredicate,PredName,Options) :-
	check_db_import_options(Options,Findall),
	db_import_1(Connection,DbPredicate,PredName,findall(Findall)).

check_db_import_options([],findall(false)):-!.
check_db_import_options([H|T],Findall):- !,
	(member(findall(Boolean),[H|T]) ->
	    Findall = Boolean ; Findall = true).
check_db_import_options(Input,_Findall):-
	domain_error(odbcImportOptionList,Input,db_import/4,4).


db_import_1(Connection,DbPredicate,M:PredName,Options) :-
	FullPredName = M:PredName,
	db_connection_check(Connection,ConnHandle),
	\+attribute(_,FullPredName,_,_,_),
	!,
	DbPredicate =.. [TableName|Attributes],
	nb_setval(col,1),
	%% --- asserts attributeX(Table,Attribute,Type,ConnHandle)
	assert_attributes_info(ConnHandle,TableName),
	functor(DbPredicate, TableName, Arity),
	%% --- asserts relation(Pred,Arity,Table,ConnHandle)
	assert(relation(FullPredName,Arity,TableName,ConnHandle)),
	%% --- asserts attribute(Num,PredNameFull,Attr,Type,ConnH)
	assert_import_attributes(FullPredName,Attributes,TableName,ConnHandle,1),
	functor(SymPred,PredName,Arity),
	db_query(Connection,M:SymPred,M:SymPred,Options).

db_import_1(_,_,PredName,_) :-
	term_to_atom(PredName, Pn),
	atomic_list_concat([Pn,' term has already been used in db_import.'], M),
	throw(error(M)).

assert_import_attributes(_,[],_,_,_).
assert_import_attributes(PredName,[HAttr|RAttr],TableName,ConnHandle,Num) :-
	(attributeX(TableName,HAttr,Type,ConnHandle)
	 ->	assert(attribute(Num,PredName,HAttr,Type,ConnHandle)),
		Num1 is Num + 1,
		assert_import_attributes(PredName,RAttr,TableName,ConnHandle,Num1)
	 ;	db_error('Table doesn''t have attribute '('Table'=TableName,
							    'Attribute'=HAttr))
	).

assert_attributes_info(ConnHandle,Table) :-
	(table_arity(ConnHandle,Table,_)
	 ->	true
	 ;	assert_attributes_info1(ConnHandle,Table),
		count_tuples(attributeX(Table,_,_,ConnHandle),Arity),
		assert(table_arity(ConnHandle,Table,Arity))
	).

assert_attributes_info1(ConnHandle,Table) :-
	nonvar(Table),
	db_get_attr(ConnHandle,Table,Col,Type),
	get_type(Type,Type1),
	\+ attributeX(Table,Col,Type1,ConnHandle),
	assert(attributeX(Table,Col,Type1,ConnHandle)),
	fail.
assert_attributes_info1(_Conn,_Table).

split_in_out(ModeTerm,BindList,InVars,InBinds,OutVars) :-
	term_to_elements(ModeTerm, _M, _, ModeList),
	length(BindList,BLL), length(InBinds,BLL),
	split_in_out_l(ModeList,BindList,InVars,InBinds,OutVars).

split_in_out_l([],_,[],_,[]).
split_in_out_l(['$bind$'(B,_Type)|ModeList],BindList,[X|InVars],
	       InBinds,[X|OutVars]) :-
	bind_inbinds(BindList,B,X,InBinds),
	split_in_out_l(ModeList,BindList,InVars,InBinds,OutVars).
split_in_out_l(['$var$'(_)|ModeList],BindList,[X|InVars],
	       InBinds,[X|OutVars]) :-
	split_in_out_l(ModeList,BindList,InVars,InBinds,OutVars).
split_in_out_l(['$const$'(C)|ModeList],BindList,[_|InVars],
	       InBinds, [C|OutVars]) :-
	split_in_out_l(ModeList,BindList,InVars,InBinds,OutVars).

bind_inbinds([],_B,_X,[]).
bind_inbinds([B1|Bs],B,X,[X1|Xs]) :-
	(B1 == B
	 ->	X = X1
	 ;	true
	),
	bind_inbinds(Bs,B,X,Xs).

:-meta_predicate db_query(:,:).
:-meta_predicate db_query(:,:,?).
:-meta_predicate db_query(?,:,:,?).
%	db_query(:Head, :Body)
%
%	Same as db_query/3 with =|Options=findall(true)|=.
%	Assert rule to intercept any call and redirect to DB.
%
db_query(Head, Body) :-
	db_query(Head,Body,_).

%%	db_query(:Head,:Body,+Options)
%
%	Same as db_query/4 with =|Connection=defaultConnection|=.
%
db_query(Head,Body,Options) :-
	db_query(defaultConnection,Head,Body,Options).

%%	db_query(+Connection,:Head,:Body,+Options)
%
%	Asserts Head predicate to Prolog local database (NOT SQL Database)
%	which when called executes SQL query to which Body is compiled.
%	See XSB documentation. 
%	It supports SQL aggregate functions (avg, min, max, count) as well as
%	=|not|= operator.
%
%	Assuming that =|card|= and =|account|= predicates
%	are created by db_import/4 the example usage 
%	that creates =|card_account|= predicate is:
%	==
%	db_query(card_account(CId,CPin,AId,AAmount),
%		(card(CId,AId,_,CPin),account(AId,AAmount))
%			),
%	==
%	Call of =|card_account|= will generate and execute proper SQL query statement
%	in above case it is SQL join statement.
%	
%	@param	Connection	Connection alias see db_update/4.
%	@param	Head		New predicate head.
%	@param	Body		Database goal.
%	@param	Options		findall(Boolean).
%
db_query(Connection,Head,Body,Options) :-
	db_connection_check(Connection,ConnHandle),
	\+ atomic(Head),
	copy_term(Head,Head1),
	assert(db_asserted_query(ConnHandle,Head1)),
	assert((Head1 :-
		db_link(reg,ConnHandle,Head,Head1,Body,Options), !,
		Head1)).

db_deallocate_query(Head) :-
	db_connection(_,ConnHandle),
	!,
	db_deallocate_query(ConnHandle,Head).

db_deallocate_query(ConnHandle,Head) :-
	retractall(db_asserted_query(ConnHandle,Head)),
	retractall(Head).

:-meta_predicate db_update(:,:).
:-meta_predicate db_update(:,:,?).
:-meta_predicate db_update(?,:,:,?).
%%% assert rule to intercept any call and redirect to DB

%%	db_update(:Head,:Body)
%
%	Same as db_update/3 with =|Options = []|=
%
db_update(Head, Body) :-
	db_update(Head,Body,[]).

%%	db_update(:Head,:Body,+Options)
%
%	Same as db_update/4 with =|Connection = defaultConnection|= see db_open/4.
%
db_update(Head,Body,Options) :-
	db_update(defaultConnection,Head,Body,Options).

%%	db_update(+Connection,:UpdateBinds,:Body,+Options)
%
%	Generates and executes update SQL statement.
%
%	Example update statement assuming that =|card_account|= 
%	and =|account|= are predicates that where created 
%	by db_import/4 or db_query/4. This statement updates 
%	=|Amount|= in record which =|Id=AccountId|=, from table 
%	represented by =|account|= predicate,
%	to =|AmountRest|= value:
%	==
%	card_account(CardId,_CardPin,AccountId,AccountAmount),
%
%	AmountRest is AccountAmount - DesiredAmount,
%
%	db_update([Amount=AmountRest],
%		(account(Id,Amount),[Id=AccountId])),
%	==
%
%	@param	Connection	Connection alias see db_open/4.
%	@param	Head		Update bindvals.
%	@param	Body		=|Body = (DatabaseGoal,WhereBindVals)|=
%						$ DatabaseGoal: data base goal
%						$ SelectBindVals: list of Var=WhereValue for select records to update.
%	@param	Options		List of options
%						$ findall(Boolean): if =|Boolean = true|= then findall mode,
%						$ rows_affected(Number): Number of rows affected by SQL update statement.
%
db_update(Connection,Head,Body,Options) :-
	(member(findall(Findall),Options) ; (Findall = false)),!,
	(member(rows_affected(Result),Options) ; true),!,

	db_connection_check(Connection,ConnHandle),
	\+ atomic(Head),

	db_flag(show_query,Backup),
	set_db_flag(show_query,off),

	copy_term((Head,Body),(Head1,Body1)),
	db_update1(Head1,Body1,Update,From),
	copy_term((Head,Body),(Head2,Body2)),
	db_update2(Head2,Body2,Where,BindVals),
	QueryTerm = query_upd(Update,From,Where),
	queries_atom([QueryTerm],SqlStmt,_),!,

	set_db_flag(show_query,Backup),

	(db_flag(show_query,on) ->sqlstmt_to_atom(SqlStmt,QAtom),writeln(QAtom);true),

    (Findall == true ->
	    (findall_db_sql_i(ConnHandle,BindVals,SqlStmt,Result))
	     ;
	     (db_sql_i(ConnHandle,BindVals,SqlStmt,Result))).

:-meta_predicate db_update1(:,:,?).
db_update1(Head,Body,Update,From):-
	copy_term((Head,Body),(Head1,Body1)),
	update_head(Head1,_Update1,Bindings1), unify_with_bindings(Bindings1),

	update_body(Body1,SelectGoal1,_SelectBind1),

	Type = reg,

	Head = M:_,
	head_to_elems(M:test, M, H, A),
	head_to_elems(M:test, M, _, Args),
	get_mode(Type,Args,Mode,A),

	ProjectionTerm1 =.. [H|Mode],
	ProjectionTerm = M:ProjectionTerm1,

	translate(ProjectionTerm,SelectGoal1,[QueryTerm],_SQLAtom,_Blist),
	QueryTerm = query(_,From,Update).

:-meta_predicate db_update2(:,:,?).
db_update2(Head,Body,Where,BindVals):-

	copy_term(Body,Body1),
	update_body(Body,SelectGoal,SelectBind,SelectBindRest),
	update_body(Body1,SelectGoal1,SelectBind1),
	unify_with_bindings(SelectBind1),

	Type = reg,

	Head = M:_,
	head_to_elems(SelectGoal, M, H, A),
	head_to_elems(SelectGoal1, M, _, Args),
	get_mode(Type,Args,Mode,A),

	ProjectionTerm1 =.. [H|Mode],
	ProjectionTerm = M:ProjectionTerm1,


	db_update2_body(SelectGoal,SelectBindRest,BodyA),

	translate(ProjectionTerm,BodyA,[QueryTerm],_SQLAtom,_),
	binding_vals(SelectBind,BindVals),

	QueryTerm = query(_,_,Where).

db_update2_body(M:SelectGoal,SelectBindRest,Body):-
	SelectBindRest = R1,
	(
	 (R1 = [])
	  -> Body = M:(SelectGoal)
	  ; (Body = M:(SelectGoal,Rest),
	                               (
					 length(R1,N),N>1,!,
					 Rest=..[','|R1]
					;
					 [Rest] = R1
					)
	    )
	).

db_update2_body_helper(_M:[],[]):-!.
db_update2_body_helper(M:[H|T],[M:H|T1]):-
	db_update2_body_helper(M:T,T1).

head_to_elems(Head,Module,PredName,Args):-
	Head = Module:Term,
	Term =.. [PredName|Args].

update_head(M:Head,M:Update,Bindings):-
	update_head1(Head,Args),
	find_bindvals(Args,Bindings,_),
	bindings_vars(Bindings,HVars),
	Update =..[update|HVars].
update_head1(Head,Args):-
	(is_list(Head), Head = Args;
	 Head=(_=_),!,Args=[Head];
	 Head =.. [_|Args]),!.

update_body(M:Body,M:WhereGoal,WhereBind):-
       update_body(M:Body,M:WhereGoal,WhereBind,_WhereBindRest).
update_body(M:Body,M:WhereGoal,WhereBind,WhereBindRest):-
	Body = (WhereGoal,WhereBind1),!,
	find_bindvals(WhereBind1,WhereBind,WhereBindRest).
update_body(M:Body,M:WhereGoal,[],[]):-
	Body = (WhereGoal).

find_bindvals(Term,[],[Term]):-
	\+ is_bindings(Term),\+ is_list(Term),!.
find_bindvals(Term,Term,[]):-
	is_bindings(Term),!.
find_bindvals([T|Ts],Bindings,Rest):-
	find_bindvals(T,B1,Rest1),
	find_bindvals(Ts,B2,Rest2),
	append(B1,B2,Bindings),
	append(Rest1,Rest2,Rest).



unify_with_bindings([]):-!.
unify_with_bindings([H=H|T]):-
	unify_with_bindings(T).

bindings_vars([],[]):-!.
bindings_vars([H=_|T],[H|R]):-
	var(H),
	bindings_vars(T,R).

binding_vals([],[]):-!.
binding_vals([_=H|T],[H|R]):-
	binding_vals(T,R).

is_bindings([_=_]).
is_bindings([_=_|T]):-
	is_bindings(T).



%%% intercepts new-moded DB call, generates and asserts new clause to
%%% (re)makes the DB call and makes the call.
db_link(Type,ConnHandle,Head,Head1,DatabaseGoal,Options) :-
	head_to_elems(Head, M, H, A),
	head_to_elems(Head1, M, _, Args),
	get_mode(Type,Args,Mode,A),

	ProjectionTerm1 =.. [H|Mode],
	ProjectionTerm = M:ProjectionTerm1,

	translate(Type,ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist),
	!,
	assert_call_rule(ProjectionTerm,Type,ConnHandle,
	                        SQLAtom,Blist,Options).

get_mode(ins,Args,Mode,A) :- !,get_mode_ins(Args,Mode,A).
get_mode(_,Args,Mode,A) :- get_mode(Args,Mode,A).

translate(reg,ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist) :-
	translate(ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist).
translate(ins,ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist) :-
	translate_ins(ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist).
translate(del,ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist) :-
	translate_del(ProjectionTerm,DatabaseGoal,_,SQLAtom,Blist).

db_transaction(Command) :-
	db_transaction(defaultConnection,Command).

% Action: commit ; rollback
db_transaction(Connection,Action) :-
	(Action = commit ; Action = rollback),
	db_connection_check(Connection,ConnHandle),
	odbc:odbc_end_transaction(ConnHandle, Action).

db_transaction(Connection, autocommit(on)) :-
	odbc:odbc_set_connect(Connection, auto_commit(true)).
db_transaction(Connection,autocommit(off)) :-
	odbc:odbc_set_connect(Connection, auto_commit(false)).


:-meta_predicate db_delete(:,:).
:-meta_predicate db_delete(?,:,:).

db_delete(Head,Body) :-
	db_delete(_Conn,Head,Body).

db_delete(Connection,Head, Body) :-
	db_connection_check(Connection,ConnHandle),
	copy_term(Head,Head1),
	assert(db_asserted_query(ConnHandle,Head1)),
	assert((Head1 :-
		db_link(del,ConnHandle,Head,Head1,Body,findall(false)), !,
		Head1)).


:-meta_predicate db_insert(:,:).
:-meta_predicate db_insert(?,:,:).

db_insert(Head, Body) :-
	db_insert(defaultConnection,Head,Body).
db_insert(Connection,Head,Body) :-
	db_connection_check(Connection,ConnHandle),
	copy_term(Head,Head1),
	assert(db_asserted_query(ConnHandle,Head1)),
	assert((Head1 :-
		db_link(ins,ConnHandle,Head,Head1,Body,findall(false)), !,
		Head1)).

assert_call_rule(ProjectionTerm,Type,ConnHandle,SQLAtom,Blist,
	            findall(Findall)) :-
	init_gensym(var),
	init_gensym(rel),
	init_gensym('BIND'),
	tokenize_term(ProjectionTerm, Tokens),!,
	term_to_elements(Tokens, M, Head, Args),
	split_in_out(Tokens,Blist,InVars,InBinds,OutVars0),
	(Type == reg -> OutVars = OutVars0 ; OutVars = _),
	term_to_elements(RuleHead, M, Head, InVars),
	converts(Args,Type,InVars,RuleBody),
	(Findall == true ->
	    asserta((RuleHead :- RuleBody, !,
		 findall_db_sql_i(ConnHandle,InBinds,SQLAtom,OutVars)))
	     ;
	     asserta((RuleHead :- RuleBody, !,
		 db_sql_i(ConnHandle,InBinds,SQLAtom,OutVars))) ).

converts([H,H2|Rest],Type,[H1|Rest1],(X, RestX)) :-
        convert(H,Type,H1,X),
        converts([H2|Rest],Type,Rest1,RestX).
converts([H],Type,[H1],X) :-
        convert(H,Type,H1,X).

convert('$var$'(_),_,A,var(A)).
convert('$bind$'(_,_),reg,A,(nonvar(A),not(A='NULL'(_X)))).
convert('$bind$'(_,_),ins,A,(nonvar(A),not(A='NULL'(_X)))).
convert('$bind$'(_,_),del,A,(nonvar(A), not((A = 'NULL'(X),var(X))))).
convert('$const$'('NULL'),reg,A,(nonvar(A),A='NULL'(_))) :- !.
convert('$const$'('NULL'),ins,A,(nonvar(A),A='NULL'(_))) :- !.
convert('$const$'('NULL'),del,A,(nonvar(A),A='NULL'(X),var(X))):- !.
convert('$const$'(C),_,A,A==C).


writelist([H]) :- writeln(H),!.
writelist([H|T]) :-
	write(H),write(' '), writelist(T).

writelistln([H]) :- writeln(H),!.
writelistln([H|T]) :-
	writeln(H), writelistln(T).

% -----------------------------------------------------------------------
db_get_schema(Command,Row) :-
	db_get_schema(defaultConnection,Command,Row).
db_get_schema(Connection,Command,Row) :-
	db_connection_check(Connection,ConnHandle),
	db_get_schema1(Command,ConnHandle,Row).

db_show_schema(Command) :-
	db_show_schema(defaultConnection,Command).
db_show_schema(Connection,Command) :-
	db_connection_check(Connection,ConnHandle),
	db_show_schema1(Command,ConnHandle).

db_get_schema1(tuples(Table),ConnHandle,Row) :-
	db_sql_i(ConnHandle,[],['SELECT * FROM ',Table], Row).

db_show_schema1(tuples(Table),ConnHandle) :-
	db_get_schema1(tuples(Table),ConnHandle,Row),
	writelist(Row), fail.
db_show_schema1(tuples(_Table),_ConnHandle).

%--------------------------------
db_get_schema1(user,ConnHandle,List) :-
        setof(A, user_tables(ConnHandle,A), List).
db_show_schema1(user,ConnHandle) :-
        db_get_schema1(user,ConnHandle,List),
        writelistln(List).

user_tables(ConnHandle,A) :-
	var(A),
	db_user_tables(ConnHandle,A).

%---------------------------
% Use 'dbo' as default user to be backwards compatible
db_get_schema1(accessible,ConnHandle,List) :-
	bagof(A, accessible_tables(ConnHandle,dbo,A), List).
db_get_schema1(accessible(User),ConnHandle, List) :-
	bagof(A, accessible_tables(ConnHandle,User,A), List).
db_get_schema1(all,ConnHandle,List) :-
	bagof([A,Type], all_tables(ConnHandle,dbo,A,Type),List).
db_get_schema1(all(User),ConnHandle, List) :-
	bagof([A,Type], all_tables(ConnHandle,User,A,Type), List).

db_show_schema1(accessible,ConnHandle) :-
	db_show_schema1(accessible(dbo),ConnHandle).
db_show_schema1(accessible(User),ConnHandle) :-
	db_get_schema1(accessible(User),ConnHandle,List),
	write(User), writeln(' owns:'), writelistln(List), fail; true.
db_show_schema1(all,ConnHandle) :-
	db_show_schema1(all(dbo),ConnHandle).
db_show_schema1(all(User),ConnHandle) :-
	db_get_schema1(all(User),ConnHandle,List),
	write(User), writeln(' owns:'), writelistln(List), fail; true.

accessible_tables(ConnHandle,User,A) :-
	all_tables(ConnHandle,User,A,TabType),
	(TabType = 'TABLE' -> true ; TabType = 'SYSTEM TABLE' -> true).

all_tables(ConnHandle,User,A,TabType) :-
	var(A),
	db_accessible_tables(ConnHandle,User,A,TabType).

%----------------------------
db_get_schema1(arity(Table),ConnHandle,N1) :-
	(table_arity(ConnHandle,Table,N1)
	 ->	true
	 ;	nb_setval(col,1),
		assert_attributes_info(ConnHandle,Table),
		table_arity(ConnHandle,Table,N1)
	).

db_show_schema1(arity(Table),ConnHandle) :-
	db_get_schema1(arity(Table),ConnHandle,N1),
	write('Table '), write( Table), write(' has arity '), writeln(N1).

%------------------------------
db_get_schema1(columns(Table),ConnHandle,List) :-
	(table_arity(ConnHandle,Table,_)
	 ->	true
	 ;	nb_setval(col,1),
		assert_attributes_info(ConnHandle,Table)
	),
	bagof(Col, Type1^attributeX(Table,Col,Type1,ConnHandle), List).

db_show_schema1(columns(Table),ConnHandle) :-
	db_get_schema1(columns(Table),ConnHandle,List),
	writelistln(List).

%--------------------------------
db_attach(Pred,Table) :-
	db_attach(defaultConnection,Pred,Table).

db_attach(Connection,Pname,table(Table)) :-
	db_connection_check(Connection,ConnHandle),
	(table_arity(ConnHandle,Table,_)
	 ->	true
	 ;	nb_setval(col, 1),
		assert_attributes_info(ConnHandle,Table)
	),
	bagof(Col, Type1^attributeX(Table,Col,Type1,ConnHandle), List),
	write('db_attach...'), writeln(List),
	DBPred =.. [Table|List],
	db_import(Connection,DBPred,Pname).

db_attach(Connection,Pname,table(Table, List)) :-
	db_connection_check(Connection,ConnHandle),
	(table_arity(ConnHandle,Table,_)
	 ->	true
	 ;	nb_setval(col, 1),
		assert_attributes_info(ConnHandle,Table)
	),
	DBPred =.. [Table|List],
	db_import(Connection,DBPred,Pname).

%--------------------------------
db_create_table(Table,Fields) :-
	db_create_table(defaultConnection,Table,Fields).

db_create_table(Connection,Table,Fields) :-
	nonvar(Table),
	nonvar(Table),
	db_sql(Connection,[],['CREATE TABLE ',Table,'(',Fields,')'], _).

%--------------------------------
db_create_index(Tname,Iname,index(_,Fields)) :-
	db_create_index(defaultConnection,Tname,Iname,index(_,Fields)).

db_create_index(Connection,Tname,Iname,index(_,Fields)) :-
	nonvar(Tname),
	nonvar(Iname),
	nonvar(Fields),
	db_sql(Connection,[],['CREATE INDEX ',Iname,' ON ',Tname,
		  ' (',Fields,' )'], _).

%--------------------------------
db_delete_table(Table) :-
	db_delete_table(defaultConnection,Table).
db_delete_table(Connection,Table) :-
	nonvar(Table),
	db_sql(Connection,[],['DROP TABLE ',Table], _).

%--------------------------------
db_delete_view(View) :-
	db_delete_view(defaultConnection,View).

db_delete_view(Connection,View) :-
	nonvar(View),
	db_sql(Connection,[],['DROP VIEW ',View], _).

%--------------------------------
db_delete_index(TableName,Index) :-
	db_delete_index(defaultConnection,TableName,Index).
db_delete_index(Connection,TableName,Index) :-
	nonvar(Index),
	db_sql(Connection,[],['DROP INDEX ',Index,' ON ',TableName],_).

count_tuples(Call,Count) :-
	nb_setval('$xsb_count',0),
	call(Call),
	nb_getval('$xsb_count', V), V1 is V + 1,
	nb_setval('$xsb_count', V1),
	fail
	;
	nb_getval('$xsb_count',Count).

%call1(X) :- call0(X).

%--------------------------------------------------
% Flags
%--------------------------------------------------

%%	db_flag(FlagName,Val)
%
%	@param	FlagName	[fail_on_error|show_query].
%	@param	Val			Depending on FlagName it can be one of:
%						$ fail_on_error:	fail, ignore, throw,
%						$ show_query:		on, off.
%
db_flag(fail_on_error,Val):-
	db_error_flg(Val).
db_flag(show_query,Val):-
	db_show_query(Val).

%%	set_db_flag(FlagName,Val)
%
%	Set FlagName flag Value (Val) see db_flag/2.
%
set_db_flag(fail_on_error,Val):-
	check_db_fail_on_error(Val),
	retractall(db_error_flg(_)),
	assert(db_error_flg(Val)).
set_db_flag(show_query,Val):-
	check_db_show_query(Val),
	retractall(db_show_query(_)),
	assert(db_show_query(Val)).

check_db_fail_on_error(Val):-
	(member(Val,[fail,ignore,throw]) ->
	    true
	 ;  domain_error(db_fail_on_error_value,Val,set_db_flag/2,2) ).

check_db_show_query(Val):-
	(member(Val,[on,off]) ->
	    true
	 ;  domain_error(db_show_query_value,Val,set_db_flag/2,2) ).

% Obsolescent -- they still work (buggily), but dont support throw.
db_flag(fail_on_error,_/*on*/,off) :-
	retractall(db_error_flg(on)), assert(db_error_flg(off)).
db_flag(fail_on_error, _/*off*/, on) :-
	retractall(db_error_flg(off)), assert(db_error_flg(on)).
db_flag(show_query, _/*off*/, on) :-
	retractall(db_show_query(off)), assert(db_show_query(on)).
db_flag(show_query, _/*on*/, off) :-
	retractall(db_show_query(on)), assert(db_show_query(off)).

%--------------------------------------------------

db_get_attr(ConnHandle,Tab,Col,Type) :-
	odbc:odbc_table_column(ConnHandle,Tab,Col,type_name(Type)).

db_accessible_tables(ConnHandle,User,Tab,TabType) :-
	odbc:odbc_current_table(ConnHandle, Tab, type(TabType)),
    odbc:odbc_current_table(ConnHandle, Tab, owner(User)).

db_user_tables(Tab) :-
	db_connection_check(defaultConnection,ConnHandle),
	db_user_tables(ConnHandle,Tab).

db_user_tables(ConnHandle,Tab) :-
    user_connection(User, ConnHandle),
    odbc:odbc_current_table(ConnHandle, Tab, owner(User)).

db_data_sources(DataSourceName,Description) :-
    odbc:odbc_data_source(DataSourceName, Description).

db_connection_exists(ConnectionName):-
        db_connection(ConnectionName,_).

db_error(ErrMsg) :-
	(db_error_flg(ignore)
	 ->	write('ERR - ODBC: '),writeln(ErrMsg)
	 ; db_error_flg(fail)
	 ->	write('ERR - ODBC: '),writeln(ErrMsg),
		fail
	 ;	%xsb_backtrace(Bt),
		throw(db_error(ErrMsg))
	).






