/*  File:    term_utils.pl
    Author:  Tomasz Maslanka
    Created: May  8 2010
*/

:-module(term_utils, [
		      atoms_to_terms/3,
		      atoms_to_terms/2,
		      unify_by_names/3,
		      binding_to_varname/2,
		      term_to_elements/4,
		      term_to_elements_meta/4
		     ]).



%%	atoms_to_terms(+ListOfAtoms, -ListOfTerms)
%
%
%

atoms_to_terms(ListOfAtoms, ListOfTerms):-
	atoms_to_terms(ListOfAtoms, ListOfTerms, _).

%%	atoms_to_terms(+ListOfAtoms, -ListOfTerms, -VarList)
%
%
%
atoms_to_terms(ListOfAtoms, ListOfTerms, VarList):-
	atoms_to_terms1(ListOfAtoms, [], ListOfTerms, VarList).

atoms_to_terms1([], LV, [], LV).
atoms_to_terms1([Atom|Atoms], LastVarList, [Term|Terms], FinalVarList) :-
	(
	( is_list(Atom) ) -> atoms_to_terms1(Atom, LastVarList, Term, V)
	                    ; atom_to_term(Atom, Term, V)
	),
	unify_by_names(LastVarList, V, VL1),
	atoms_to_terms1(Atoms, VL1, Terms, FinalVarList).
% ------------------------------------------------------------------------







%%	unify_by_names(+Binding1, +Binding2, -Unified)
%
unify_by_names(Binding1, Binding2, Unified):-
	merge_list(Binding1, Binding2, Unified).

merge_list([], V2, V2).

merge_list([V1|V1s], V2, V3):-
	merge_list1(V1, V2, Vm),
	merge_list(V1s, Vm, V3).

merge_list1(H, [H|T], [H|T]):-!.
merge_list1(Elem, [], [Elem]):-!.
merge_list1(Elem, [H|T], [H|Mt]) :-
	Elem \= H,
	merge_list1(Elem, T, Mt).

% ------------------------------------------------------------------------






%%	binding_to_varname(?Binding, ?VarName)
%
binding_to_varname(N=_, N):-!.

%%	var_names(?BindingList, ?VarNames)
%
binding_to_varname([], []).
binding_to_varname([Bind|Binds], [Name|Names]) :-
	binding_to_varname(Bind, Name),
	binding_to_varname(Binds, Names).



%%	term_to_elements(?Term, ?M, ?Functor, ?Args)
%
%

term_to_elements(Term, M, Term, []):-
	atomic(Term),!,
	strip_module(Term,M,_Plain).

term_to_elements(Term, M, Functor, Args):-
	compound(Term),
	strip_module(Term, M, Plain),!,
	Plain=..[Functor|Args].

term_to_elements(Term, M, Functor, Args):-
	atomic(Functor), nonvar(Args),
	Term1 =.. [Functor|Args],
	Term = M:Term1.


:-meta_predicate term_to_elements_meta(:,?,?,?).

term_to_elements_meta(Term, M, Functor, Args):-
	term_to_elements(Term, M, Functor, Args).














